"""
This module provides URL parsing and query string handling for the search functionality.
"""

from typing import Any

from . import EnvPlugin, ConfigDict, register_plugin, convert_values

SEARCH_SCHEMES = {
    "elasticsearch": "haystack.backends.elasticsearch_backend.ElasticsearchSearchEngine",
    "elasticsearch2": "haystack.backends.elasticsearch2_backend.Elasticsearch2SearchEngine",
    "elasticsearch+dsl": None,
    "elasticsearch-dsl": None,
    "solr": "haystack.backends.solr_backend.SolrEngine",
    "whoosh": "haystack.backends.whoosh_backend.WhooshEngine",
    "xapian": "haystack.backends.xapian_backend.XapianEngine",
    "simple": "haystack.backends.simple_backend.SimpleEngine",
}


@register_plugin("search_url")
class SearchPlugin(EnvPlugin):
    """
    Plugin for handling search configuration
    """

    VAR = "SEARCH_URL"
    CONTEXTS = ["search"]

    def get_backend(self, url: str, **kwargs: Any) -> object:  # noqa: C901
        parsed = self.parse_url(url, context=self.CONTEXTS)
        engine = kwargs.get("engine", None)
        options = kwargs.get("options", {})
        config = ConfigDict()

        url_scheme = self.DEFAULT_SCHEME
        if not parsed.scheme:
            raise ValueError("Missing search scheme or url parse error")
        try:
            config["BACKEND"] = engine or SEARCH_SCHEMES[parsed.scheme]
        except KeyError as e:
            raise ValueError(f"Unknown search scheme: {parsed.scheme}") from e

        if parsed.qs:
            options.update(parsed.qs)
            url_scheme = options.pop("SCHEME", url_scheme)
        convert_values(options)
        # do some conversions
        if "EXCLUDED_INDEXES" in options:
            config["EXCLUDED_INDEXES"] = options.pop("EXCLUDED_INDEXES").split(",")
        config["INCLUDE_SPELLING"] = options.pop("INCLUDE_SPELLING", None)
        if "BATCH_SIZE" in options:
            config["BATCH_SIZE"] = options.pop("BATCH_SIZE")

        if parsed.path:
            path = parsed.path[1:]
            if path and path.endswith("/"):
                path = path[:-1]
            config["NAME"] = path

        config["KWARGS"] = options.get("KWARGS", None)
        config["TIMEOUT"] = options.get("TIMEOUT", None)
        config["INDEX_NAME"] = options.get("INDEX_NAME", None)
        match parsed.scheme:
            case "elasticsearch+dsl" | "elasticsearch-dsl":
                config["hosts"] = parsed.to_url(scheme=url_scheme)
            case "elasticsearch" | "elasticsearch2":
                if "NAME" in config:
                    # get index from the last segment of config["NAME"]
                    config["INDEX_NAME"] = config["NAME"].split("/")[-1]
                    # and remove it from config["NAME"]
                    config["NAME"] = "/".join(config["NAME"].split("/")[:-1])
            case "whoosh":
                config["STORAGE"] = options.get("STORAGE", None)
                config["POST_LIMIT"] = options.get("POST_LIMIT", None)
                config["NAME"] = f"/{config['NAME']}"
            case "xapian":
                config["FLAGS"] = options.get("FLAGS", None)
            case _:
                pass
        config["URL"] = parsed.to_url(scheme=url_scheme, name=None)

        return config



import re

from . import EnvPlugin, ConfigDict, register_plugin

POSTGRES_ENGINE = "django.db.backends.postgresql"
MYSQL_ENGINE = "django.db.backends.mysql"
SQLITE_ENGINE = "django.db.backends.sqlite3"

DB_ENGINES = {
    "postgres": POSTGRES_ENGINE,
    "postgresql": POSTGRES_ENGINE,
    "psql": POSTGRES_ENGINE,
    "pgsql": POSTGRES_ENGINE,
    "postgis": "django.contrib.gis.db.backends.postgis",
    "mysql": MYSQL_ENGINE,
    "mysql2": MYSQL_ENGINE,
    "mysql-connector": "mysql.connector.django",
    "mysqlgis": "django.contrib.gis.db.backends.mysql",
    "mssql": "sql_server.pyodbc",
    "oracle": "django.db.backends.oracle",
    "pyodbc": "sql_server.pyodbc",
    "redshift": "django_redshift_backend",
    "spatialite": "django.contrib.gis.db.backends.spatialite",
    "sqlite": SQLITE_ENGINE,
    "ldap": "ldapdb.backends.ldap",
}

# DB_OPTIONS = [
#     "CONN_MAX_AGE",
#     "ATOMIC_REQUESTS",
#     "AUTOCOMMIT",
#     "SSLMODE",
#     "SSLROOTCERT",
#     "TEST",
#     # extensions
#     "READ_ONLY",
#     "READONLY",
#     "HTTP_METHODS",
#     "HTTP_WRITE_PATHS",
#     "HTTP_WRITE_STICKY",
# ]


def is_postgres(engine):
    return engine in (
        POSTGRES_ENGINE,
        "django.db.backends.postgresql_psycopg2",
        "django.contrib.gis.db.backends.postgis",
        "django_redshift_backend",
    )


@register_plugin("database_url")
class DatabasePlugin(EnvPlugin):
    """
    Plugin for handling database configuration
    """

    VAR = "DATABASE_URL"
    CONTEXTS = ["database"]

    def get_backend(self, url: str, **kwargs) -> object:
        # sourcery skip: extract-method
        parsed = self.parse_url(url, context=self.CONTEXTS)
        backend = kwargs.get("backend", None)
        options = kwargs.get("options", {})
        config = ConfigDict()
        if parsed.scheme == "sqlite":
            config["NAME"] = (
                ":memory:"
                if not parsed.path or re.match(r"/:?memory:?", parsed.path)
                else parsed.path
            )
            config["ENGINE"] = backend or SQLITE_ENGINE
        else:
            try:
                config["ENGINE"] = backend or DB_ENGINES[parsed.scheme]
            except KeyError as e:
                raise ValueError(f"Unsupported database scheme: {parsed.scheme}") from e
            config["NAME"] = parsed.path[1:] if parsed.path else parsed.path
            config["HOST"] = parsed.hostname
            config["PORT"] = parsed.port or None
            config["USER"] = parsed.username
            config["PASSWORD"] = parsed.password
        if parsed.qs:
            options |= parsed.qs
        if options:
            if schema := options.pop("currentSchema", None):
                if is_postgres(config["ENGINE"]):
                    options["options"] = f"-c search_path={schema}"
            config["OPTIONS"] = options
        return config



from django.utils.version import get_complete_version

from . import EnvPlugin, ConfigDict, register_plugin, convert_values

DJANGO_VERSION = get_complete_version()

REDIS_CACHE_BACKEND = (
    "django_redis.cache.RedisCache"
    if DJANGO_VERSION[0] < 4
    else "django.core.cache.backends.redis.RedisCache"
)
DJANGO_LOCMEM_CACHE_BACKEND = "django.core.cache.backends.locmem.LocMemCache"
CACHE_SCHEMES = {
    "dbcache": "django.core.cache.backends.db.DatabaseCache",
    "dummycache": "django.core.cache.backends.dummy.DummyCache",
    "filecache": "django.core.cache.backends.filebased.FileBasedCache",
    "locmem": DJANGO_LOCMEM_CACHE_BACKEND,
    "locmemcache": DJANGO_LOCMEM_CACHE_BACKEND,
    "memcache": "django.core.cache.backends.memcached.MemcachedCache",
    "pymemcache": "django.core.cache.backends.memcached.PyLibMCCache",
    "rediscache": REDIS_CACHE_BACKEND,
    "redis+socket": REDIS_CACHE_BACKEND,
    "redis": REDIS_CACHE_BACKEND,
    "rediss": REDIS_CACHE_BACKEND,
}


@register_plugin("cache_url")
class CachePlugin(EnvPlugin):
    """
    Plugin for handling database configuration
    """

    VAR = "CACHE_URL"
    CONTEXTS = ["caches"]

    def get_backend(self, url: str, **kwargs) -> object:  # noqa: C901
        parsed = self.parse_url(url, context=self.CONTEXTS)
        backend = kwargs.get("backend", None)
        options = kwargs.get("options", {})
        config = ConfigDict()

        if not parsed.scheme:
            raise ValueError("Missing cache scheme or url parse error")
        try:
            config["BACKEND"] = backend or CACHE_SCHEMES[parsed.scheme]
        except KeyError as e:
            raise ValueError(f"Unknown cache scheme: {parsed.scheme}") from e

        host = parsed.hostname
        name = parsed.path
        match parsed.scheme:
            case "filecache":
                config["LOCATION"] = f"{host}{name}"
            case "redis" | "rediscache" | "pymemcache" | "redis+socket":
                # note: SSL is not supported for unix domain sockets, so "rediss" is not here
                if parsed.hostname == "unix" or not parsed.hostname:
                    path = name or "/tmp/redis.sock"
                    config["LOCATION"] = f"unix://{path}"
                else:
                    config["LOCATION"] = parsed.to_url()
            case "dummycache":
                config["LOCATION"] = config["NAME"] = None
            case "dbcache":
                config["LOCATION"] = parsed.path[1:] if parsed.path else ""
            case "memcache":
                config["LOCATION"] = parsed.to_url(scheme=None)
            case _:
                config["LOCATION"] = parsed.to_url()
        if parsed.qs:
            options.update(parsed.qs)
        convert_values(options)
        if options:
            config["OPTIONS"] = options
        # special handling of these
        config["TIMEOUT"] = options.pop("timeout", None)
        config["KEY_PREFIX"] = options.pop("key_prefix", None)
        return config



# -*- coding: utf-8 -*-
"""
Wrapper around os.environ with django config value parsers
"""

import contextlib
import inspect
import importlib
from typing import List

from django.core.exceptions import ImproperlyConfigured
from envex import Env

_DEFAULT_PREFIX = "DJANGO_"
_USE_DEFAULT_PREFIX = object()


class DjangoEnv(Env):
    """
    Wrapper around os.environ with .env enhancement django, and Hashicorp vault support
    """

    def __init__(self, *args, **kwargs):
        """
        @param args: dict (optional) environment variables
        @param prefix: (optional) str prefix to add to all variables (default=DJANGO_)
        @param exception: (optional) Exception class to raise on error (default=KeyError)
        @param environ: dict | None default base environment (os.environ is default)
        @param readenv: bool If True, values will be read from .env files (default=**True**)
        - if True: the following additional args may be used
        @param env_file: str name of the environment file (.env or $ENV default)
        @param search_path: Union[List[str], List[Path]], str] path(s) to search for env_file
        @param overwrite: bool whether to overwrite existing environment variables (default=False)
        @param parents: bool whether to search parent directories for env_file (default=**True**)
        @param update: bool whether to update os.environ with values from env_file (default=False)
        @param errors: bool whether to raise error on missing env_file (default=False)
        - kwargs for vault/secrets manager:
        @param url: (optional) vault url, default is $VAULT_ADDR
        @param token: (optional) vault token, default is $VAULT_TOKEN or ~/.vault-token
        @param cert: (optional) tuple (cert, key) or str path to client cert/key files
        @param verify: (optional) bool | str whether to verify server cert or set ca cert path (default=True)
        @param cache_enabled: (optional) bool whether to cache secrets (default=True)
        @param base_path: (optional) str base path for secrets (default=None)
        @param engine: (optional) str vault secrets engine (default=None)
        @param mount_point: (optional) str vault secrets mount point (default=None, determined by engine)
        @param timeout: (optional) int timeout for connecting to vault (default=5)
        @param working_dirs: (optional) bool whether to include PWD/CWD (default=True)
        -
        @param kwargs: (optional) environment variables to add/override
        """
        self.prefix = kwargs.pop("prefix", _DEFAULT_PREFIX)
        # by default, use Django config exception in preference to KeyError
        kwargs.setdefault("exception", ImproperlyConfigured)
        # change default to read .env files and search parents as well
        kwargs.setdefault("readenv", True)
        kwargs.setdefault("parents", True)
        super().__init__(*args, **kwargs)
        self._init_plugins()

    def _with_prefix(self, var, prefix):
        if prefix is _USE_DEFAULT_PREFIX:
            prefix = self.prefix
        if (
            var
            and prefix
            and not (isinstance(var, str) and var.startswith(prefix))
            and not super().is_set(var)
        ):
            var = f"{prefix}{var}"
        return var

    def unset(self, var, prefix=_USE_DEFAULT_PREFIX):
        super().unset(self._with_prefix(var, prefix=prefix))

    def is_set(self, var, prefix=_USE_DEFAULT_PREFIX):
        return super().is_set(self._with_prefix(var, prefix=prefix))

    def is_all_set(
        self, *_vars: str | List[str | list | tuple], prefix=_USE_DEFAULT_PREFIX
    ):
        def check_nested(v):
            if isinstance(v, (list, tuple)):
                return all(check_nested(item) for item in v)
            return self.is_set(self._with_prefix(v, prefix=prefix))

        return all(check_nested(v) for v in _vars)

    def is_any_set(
        self, *_vars: str | List[str | list | tuple], prefix=_USE_DEFAULT_PREFIX
    ):
        def check_nested(v):
            if isinstance(v, (list, tuple)):
                return any(check_nested(item) for item in v)
            return self.is_set(v, prefix=prefix)

        return any(check_nested(v) for v in _vars)

    def get(self, var, default=None, prefix=_USE_DEFAULT_PREFIX):
        return super().get(self._with_prefix(var, prefix=prefix), default=default)

    def int(self, var, default=None, prefix=_USE_DEFAULT_PREFIX) -> int:
        val = self.get(var, default=default, prefix=prefix)
        return self._int(val)

    def float(self, var, default=None, prefix=_USE_DEFAULT_PREFIX) -> float:
        val = self.get(var, default=default, prefix=prefix)
        return self._float(val)

    def bool(self, var, default=None, prefix=_USE_DEFAULT_PREFIX) -> bool:
        val = self.get(var, default=default, prefix=prefix)
        return bool(val) if isinstance(val, (bool, int)) else self.is_true(val)

    def list(self, var, default=None, prefix=_USE_DEFAULT_PREFIX) -> list:
        val = self.get(var, default=default, prefix=prefix)
        return val if isinstance(val, (list, tuple)) else self._list(val)

    def check_var(self, var, default=None, prefix=_USE_DEFAULT_PREFIX, raise_error=True):
        """
        Override variable name to insert prefix unless the raw var is set
        """
        var = self._with_prefix(var, prefix=prefix)
        return super().check_var(var, default=default, raise_error=raise_error)

    django_env_typemap = {
        "int": int,
        "bool": bool,
        "float": float,
        "list": list,
    }

    # noinspection PyShadowingBuiltins
    def __call__(self, var=None, default=None, **kwargs):
        prefix = kwargs.pop("prefix", _USE_DEFAULT_PREFIX)
        # This is tied to django-class-settings (optional dependency), which allows
        # omitting the 'name' parameter and using the setting name instead
        if var is None:
            _kwargs = kwargs | {
                "name": var,
                "prefix": prefix if prefix is _USE_DEFAULT_PREFIX else self.prefix,
                "default": default,
            }
            # try using class_settings version if installed
            with contextlib.suppress(ImportError):
                # noinspection PyUnresolvedReferences
                from class_settings.env import DeferredEnv

                return DeferredEnv(
                    self, kwargs=_kwargs, optional=kwargs.get("optional", False)
                )
            # otherwise, use our own implementation (handles module level vars)
            from .deferred import DeferredSetting

            scope = inspect.currentframe().f_back

            # class settings not installed
            return DeferredSetting(env=self, scope=scope, kwargs=_kwargs)

        # set to the provided default if not already set
        if (
            default is not None
            and not self.is_set(var)
            and not self.is_set(self._with_prefix(var, prefix=prefix))
        ):
            self.set(var, default)
        # resolve entry point by the type
        with contextlib.suppress(KeyError):
            _type = kwargs.pop("type", "str")
            _type = _type if isinstance(_type, str) else _type.__name__
            return self.django_env_typemap[_type](
                self, var, default=default, prefix=prefix
            )
        return self.get(var, default=default, prefix=prefix)

    @staticmethod
    def _init_plugins():
        """
        Initialize all plugins in the plugin package
        :return:
        """
        from . import plugin

        # noinspection PyProtectedMember
        if not plugin._initialized:
            import os

            package_path = plugin.__path__[0]  # Use the first path in the list
            package_prefix = plugin.__name__
            for filename in os.listdir(package_path):
                if filename.endswith(".py") and not filename.startswith("_"):
                    module_name = filename[:-3]  # Remove the .py extension
                    module_name = f"{package_prefix}.{module_name}"
                    spec = importlib.util.find_spec(module_name)
                    if spec is not None:
                        importlib.import_module(module_name)

    def __getattr__(self, name):
        """
        Handle calls to unknown methods by checking against registered plugins.
        """
        from . import plugin

        rplugin = plugin.get_plugin_from_name(name)
        if rplugin is not None:
            # virtual handler for plugins
            def handler(
                var=None,
                *,
                default=None,
                backend=None,
                engine=None,
                prefix=_USE_DEFAULT_PREFIX,
                **kwargs,
            ):
                if backend and engine:
                    raise ValueError("You cannot specify both 'backend' and 'engine'")
                if var is None:
                    # Use the default value specified by the plugin if necessary
                    var = getattr(rplugin, "VAR", None)
                # Determine the URL using the check_var method
                url = self.check_var(var, prefix=prefix, default=default)
                # Call the registered plugin handler with the resolved arguments
                kwargs["backend"] = backend
                kwargs["engine"] = engine
                kwargs = {k: v for k, v in kwargs.items() if v is not None}
                return rplugin.get_backend(url, **kwargs)

            # Cache the handler for future calls
            self.__dict__[name] = handler
            return handler

        # If no plugin matches, fallback to normal attribute handling
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )



# -*- coding: utf-8 -*-
from functools import wraps

from django.conf import LazySettings


def cache_setting(self, name, value):
    if isinstance(value, DeferredSetting):
        if not hasattr(cache_setting, "cache"):
            cache_setting.cache = {}
        cache = cache_setting.cache
        if name not in cache:
            cache[name] = value.setting(name)
        value = cache[name]
    return value


def deferred_getattr(func):
    # The @wraps decorator is used to preserve the original function's metadata
    @wraps(func)
    def wrapper(self, name):
        return cache_setting(self, name, func(self, name))

    return wrapper


def deferred_getattribute(func):
    @wraps(func)
    def wrapper(self, name):
        return cache_setting(self, name, func(self, name))

    return wrapper


def deferred_handler():
    # at least one DeferredSetting is being used, so override the __getattr__ handler for the setting module
    # to catch any DeferredSetting use and return an appropriate value from the environment
    if not getattr(deferred_handler, "__enabled__", False):
        # need to overwrite both because __getattr__ is not called if the setting is defined (no longer lazy)
        setattr(LazySettings, "__getattr__", deferred_getattr(LazySettings.__getattr__))
        setattr(
            LazySettings,
            "__getattribute__",
            deferred_getattribute(LazySettings.__getattribute__),
        )
        deferred_handler.__enabled__ = True


class DeferredSetting:
    # similar to django-class-settings DeferredEnv but simpler
    # and handles settings at module level not in a Settings class

    def __init__(self, env, *, scope, kwargs):
        self._name = kwargs.pop("name", None) or None
        self._env, self._scope, self._kwargs = env, scope, kwargs
        deferred_handler()

    def __get_variable_name(self, name):
        name = self._name or name
        if name is None:
            names = [n for n, v in self._scope.f_locals.items() if v is self]
            name = names[0] if names else None
        return name

    def setting(self, name):
        name = self.__get_variable_name(name)
        return self._env.get(name, **self._kwargs) if name else ""

    def __repr__(self):
        return self.setting(self._name) or ""



from . import EnvPlugin, ConfigDict, register_plugin

EMAIL_AMAZON_SES = "django_ses.SESBackend"
EMAIL_SMTP = "django.core.mail.backends.smtp.EmailBackend"
EMAIL_SCHEMES = {
    "smtp": EMAIL_SMTP,
    "smtps": EMAIL_SMTP,
    "smtp+tls": EMAIL_SMTP,
    "smtp+ssl": EMAIL_SMTP,
    "consolemail": "django.core.mail.backends.console.EmailBackend",
    "filemail": "django.core.mail.backends.filebased.EmailBackend",
    "memorymail": "django.core.mail.backends.locmem.EmailBackend",
    "dummymail": "django.core.mail.backends.dummy.EmailBackend",
    "amazonses": EMAIL_AMAZON_SES,
    "amazon-ses": EMAIL_AMAZON_SES,
}


@register_plugin("email_url")
class EmailPlugin(EnvPlugin):
    """
    Plugin for handling email configuration
    """

    VAR = "EMAIL_URL"
    CONTEXTS = ["email"]

    def get_backend(self, url: str, **kwargs) -> object:
        parsed = self.parse_url(url, context=self.CONTEXTS)
        backend = kwargs.get("backend", None)
        options = kwargs.get("options", {})
        config = ConfigDict()

        if not parsed.scheme:
            raise ValueError("Missing email scheme or url parse error")
        try:
            config["EMAIL_BACKEND"] = backend or EMAIL_SCHEMES[parsed.scheme]
        except KeyError as e:
            raise ValueError(f"Unknown email scheme: {parsed.scheme}") from e

        config.update(options)
        default_port = 25
        if parsed.scheme in {"smtps", "smtp+tls"}:
            config["EMAIL_USE_TLS"] = True
            default_port = 587
        elif parsed.scheme == "smtp+ssl":
            config["EMAIL_USE_SSL"] = True
            default_port = 465

        config.update(
            {
                "EMAIL_FILE_PATH": parsed.path[1:] if parsed.path else None,
                "EMAIL_HOST_USER": parsed.username,
                "EMAIL_HOST_PASSWORD": parsed.password,
                "EMAIL_HOST": parsed.hostname or "localhost",
                "EMAIL_PORT": int(parsed.port) if parsed.port else default_port,
            }
        )
        if parsed.qs:
            config.update(parsed.qs)

        # if update=globals() can update settings module directly
        # if update=settings_class.__dict__ if using class-based settings
        if (update := kwargs.get("update", None)) is not None:
            if isinstance(update, dict):
                update.update(config)

        return config



from typing import Optional
from dataclasses import dataclass
from yarl import URL


__all__ = (
    "URLParser",
    "ParsedUrl",
    "default_parser",
)

_none = object()


@dataclass
class ParsedUrl:
    scheme: str
    username: Optional[str] = None
    password: Optional[str] = None
    hostname: Optional[str] = None
    port: Optional[str | int] = None
    path: Optional[str] = None
    qs: Optional[dict] = None

    def to_url(
        self,
        scheme: Optional[str] = _none,
        name: Optional[str] = _none,
        port: Optional[str] = _none,
        qs: bool = False,
    ) -> str:
        scheme = scheme if scheme is not _none else self.scheme or "https"
        scheme = f"{scheme}://" if scheme else ""
        if self.username:
            userpass = (
                f"{self.username}:{self.password}@"
                if self.password
                else f"{self.username}@"
            )
        elif self.password:
            userpass = f":{self.password}@"
        else:
            userpass = ""
        url = f"{scheme}{userpass}{self.hostname or ''}"
        port = port if port is not _none else self.port
        if port:
            url += f":{port}"
        name = name if name is not _none else self.path
        if name:
            url += name
        if self.qs and qs:
            url += "?" + "&".join(f"{k}={v}" for k, v in self.qs.items())
        return url

    def __str__(self) -> str:
        return f"<ParsedUrl: {self.to_url()}"


class URLParser:
    """
    Base class for URL parsers. Subclasses must implement the __call__ method to parse URLs.
    Parsers normalise the given URL and return ParsedUrl objects.
    """

    DEFAULT_SCHEME = None

    @property
    def default_scheme(self) -> str:
        """
        Returns the default scheme for the parser.

        Returns:
            The default scheme.
        """
        return self.DEFAULT_SCHEME or "https"

    def __call__(self, url: str, **kwargs) -> ParsedUrl:
        """
        Parses the given URL and returns a ParsedUrl object.

        Args:
            url: The URL to be parsed.
            kwargs: Additional arguments for the parser.

        Returns:
            A ParsedUrl object
        Throws:
            ValueError: If the URL can't be parsed.
        """
        if "://" not in url and "/" in url:
            url = f"https://{url}"
        parsed = URL(url)
        if parsed.scheme or parsed.host or parsed.path:
            return ParsedUrl(
                scheme=parsed.scheme,
                username=parsed.user,
                password=parsed.password,
                hostname=parsed.host,
                port=parsed.port,
                path=parsed.path,
                qs=parsed.query or None,
            )
        else:
            raise ValueError(f"Invalid URL: {url}")


default_parser = URLParser()



from django.utils.version import get_complete_version

from . import EnvPlugin, ConfigDict, register_plugin, convert_values

DJANGO_VERSION = get_complete_version()

MODULE_PREFIX = (
    "django_tasks"
    if (DJANGO_VERSION[0] < 5 > DJANGO_VERSION[0]) or DJANGO_VERSION[1] < 2
    else "django.core.tasks"
)
REDIS_QUEUE_BACKEND = f"{MODULE_PREFIX}.backends.RedisQueue"
REDIS_BACKEND = f"{MODULE_PREFIX}.backends.RedisBackend"
DATABASE_BACKEND = f"{MODULE_PREFIX}.backends.database.DatabaseBackend"

TASKS_SCHEMES = {
    "redis": REDIS_BACKEND,
    "redis-queue": REDIS_QUEUE_BACKEND,
    "postgres": DATABASE_BACKEND,
    "postgresql": DATABASE_BACKEND,
    "mysql": DATABASE_BACKEND,
    "sqlite": DATABASE_BACKEND,
    "dummy": f"{MODULE_PREFIX}.backends.dummy.DummyBackend",
    "immediate": f"{MODULE_PREFIX}.backends.immediate.ImmediateBackend",
}


@register_plugin("tasks_url")
class TasksPlugin(EnvPlugin):
    """
    Plugin for handling database configuration
    """

    VAR = "CACHE_URL"
    CONTEXTS = ["caches"]

    def get_backend(self, url: str, **kwargs) -> object:
        parsed = self.parse_url(url, context=self.CONTEXTS)
        backend = kwargs.get("backend", None)
        options = ConfigDict(kwargs.get("options", {}))
        config = ConfigDict()

        if not parsed.scheme:
            raise ValueError("Missing tasks scheme or url parse error")
        try:
            config["BACKEND"] = backend or TASKS_SCHEMES[parsed.scheme]
        except KeyError as e:
            raise ValueError(f"Unknown tasks scheme: {parsed.scheme}") from e

        url_scheme = parsed.scheme
        name = parsed.path[1:] if parsed.path else None
        match parsed.scheme:
            case "redis" | "redis-queue":
                if parsed.hostname == "unix":
                    path = name or "tmp/redis.sock"
                    config["URL"] = f"unix:///{path}"
                else:
                    url_scheme = "redis"
            case "postgres" | "postgresql" | "mysql" | "sqlite":
                # which db to use
                config["DATABASE"] = parsed.path[1:] if parsed.path else "default"
                config["URL"] = parsed.to_url()
            case "dummy" | "immediate":
                pass  # no additional options

        config["URL"] = parsed.to_url(scheme=url_scheme)
        if parsed.qs:
            options.update(parsed.qs)
        convert_values(options)
        config["ENQUEUE_ON_COMMIT"] = options.pop("ENQUEUE_ON_COMMIT", None)
        config["QUEUES"] = options.pop("QUEUES", None)
        if options:
            config["BACKEND_OPTIONS"] = options
        return config



import contextlib
from abc import ABC, abstractmethod
from typing import Dict, Type, Any
import threading

__all__ = ("EnvPlugin", "ConfigDict", "register_plugin")

from ..parser import ParsedUrl, default_parser


class ConfigDict(dict):
    def __setitem__(self, key, value):
        if value:
            super().__setitem__(key, value)
        elif key in self:
            super().__delitem__(key)

    def update(self, *args, **kwargs):
        for key, value in dict(*args, **kwargs).items():
            self[key] = value
        return self


def is_int(value: Any) -> bool:
    if value:
        if not isinstance(value, str):
            value = str(value)
        with contextlib.suppress(ValueError):
            int(value)
            return True
    return False


def is_float(value: Any) -> bool:
    if value:
        if not isinstance(value, str):
            value = str(value)
        with contextlib.suppress(ValueError):
            float(value)
            return True
    return False


def is_true(value: Any) -> bool:
    if not value:
        return False
    if isinstance(value, str):
        value = value.lower()
        return any(value.startswith(x) for x in ("tr", "ye", "1", "en"))
    return bool(value)


def convert_values(options: dict):
    # convert numerical values to numbers or booleans
    for k, v in options.items():
        if is_int(v):
            options[k] = int(v)
        elif is_float(v):
            options[k] = float(v)


class EnvPlugin(ABC):
    DEFAULT_SCHEME = "https"

    @property
    def parser(self):
        return default_parser

    def parse_url(self, url: str, **kwargs) -> ParsedUrl:
        return self.parser(url, **kwargs)

    @abstractmethod
    def get_backend(self, url: str, **kwargs) -> object:
        # Return a Django object or dictionary
        pass


_initialized = False
_registered_plugins: Dict = {}
_registered_plugins_lock = threading.Lock()


def _register_plugin(plugin_cls: type[EnvPlugin], name: str):
    if not issubclass(plugin_cls, EnvPlugin):
        raise TypeError(f"{plugin_cls.__name__} must implement EnvPlugin")
    with _registered_plugins_lock:
        global _initialized
        _registered_plugins[name] = plugin_cls()
        _initialized = True


def register_plugin(name: str):
    """
    Class decorator to automatically register a plugin subclass in the plugin registry.

    Args:
        name: DjangoEnv method name to handle

    Raises:
        TypeError: If the decorated class doesn't implement the EnvPlugin protocol.
    """

    def decorator(plugin_cls: Type[EnvPlugin] | str):
        _register_plugin(plugin_cls, name)
        return plugin_cls

    return decorator  # Return the parser class itself


def get_plugin_from_name(name: str) -> EnvPlugin:
    with _registered_plugins_lock:
        return _registered_plugins.get(name)



# -*- coding: utf-8 -*-
from envex import dot_env
from envex.dot_env import load_dotenv, load_env

from .env_django import DjangoEnv as Env

__all__ = (
    "dot_env",
    "load_env",
    "load_dotenv",
    "Env",
)
